#!/bin/sh
#
# This script should install all the binaries and scripts with proper
# ownership and permissions, creating directories (with proper owner
# and permissions) as needed.  It should only install config files when
# there is no exisiting config file in the installation path.
#
# It needs to be run from the top of the source tree.
#

. ./shoki.conf

bins='sensor/sensor parser/parser alert/alert session/session service/service fp/fp rt/rt scripts/check_sensor scripts/query_pickup scripts/check_current.pl scripts/send_dump.pl scripts/sorter scripts/collector scripts/collector.add_data scripts/alerter'
confs='parser_filterlist.conf parser_arachnids.conf parser_backdoors.conf parser_icmp.conf parser_ip.conf parser_misc.conf parser_overflow.conf parser_snort.conf parser_tcp.conf parser_udp.conf parser_web.conf'
etcs='sensor.conf service.conf session.conf alert.conf parser.conf service_filterlist.conf session_filterlist.conf sensor_filter fp_filter rt_filter shoki.conf collector.conf collector_sensors.conf meshuga.conf meshuga_filterlist.conf fp.conf rt.conf sensor_filter meshuga_filter alert_filter alert_filterlist.conf passive_fingerprints ip_filterlist.conf alert_variables.conf utils.pl db_utils.pl'
mans='man5/filterlist.conf.5 man6/meshuga.6 man1/parser.1 man8/sensor.8 man1/service.1 man1/session.1 man5/shoki.conf.5 man5/shoki_configs.5'
filters=`ls -A ./etc/filters`

make_dir()
{
	if [ ! -d $1 ]; then
		mkdir -p -m $2 $1
		chown $3:$4 $1
	fi
}

make_dir ${INSTALL_PREFIX} 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${ETCDIR} 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${CONFDIR} 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${BINDIR} 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/old 0770 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/pickup 0770 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/log 0770 root ${IDSGRP}
make_dir ${INSTALL_PREFIX}/central 0770 root ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${ETCDIR}/filters 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${MANDIR} 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${MANDIR}/man1 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${MANDIR}/man5 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${MANDIR}/man6 0550 ${IDSUSER} ${IDSGRP}
make_dir ${INSTALL_PREFIX}/${MANDIR}/man8 0550 ${IDSUSER} ${IDSGRP}
foo=`ls -ld ${INSTALL_PREFIX}/log | awk '{print $4}'`
if [ ${foo} = ${IDSGRP} ]; then
	chmod g+s ${INSTALL_PREFIX}/log
fi
touch ${INSTALL_PREFIX}/${ETCDIR}/pending
chown ${IDSUSER}:${IDSGRP} ${INSTALL_PREFIX}/${ETCDIR}/pending
chmod 0660 ${INSTALL_PREFIX}/${ETCDIR}/pending
for foo in ${bins} ; do
	echo "Installing ${foo} in ${INSTALL_PREFIX}/${BINDIR}"
	install -c -o root -g ${IDSGRP} -m 550 ${foo} ${INSTALL_PREFIX}/${BINDIR}
done
if [ -f meshuga/meshuga ]; then
	echo "Installing meshuga/meshuga in ${INSTALL_PREFIX}/${BINDIR}"
	install -c -o root -g ${IDSGRP} -m 550 meshuga/meshuga ${INSTALL_PREFIX}/${BINDIR}/meshuga
fi
for foo in ${etcs} ; do
	if [ ! -f ${INSTALL_PREFIX}/${ETCDIR}/${foo} ]; then
		echo "Installing etc/${foo} in ${INSTALL_PREFIX}/${ETCDIR}"
		install -c -o root -g ${IDSGRP} -m 440 etc/${foo} ${INSTALL_PREFIX}/${ETCDIR}/${foo}
	fi
done
for foo in ${confs} ; do
	if [ ! -f ${INSTALL_PREFIX}/${ETCDIR}/${foo} ]; then
		echo "Installing etc/${foo} in ${INSTALL_PREFIX}/${CONFDIR}"
		install -c -o root -g ${IDSGRP} -m 440 etc/${foo} ${INSTALL_PREFIX}/${CONFDIR}/${foo}
	fi
done
for foo in ${mans} ; do
	echo "Installing man/${foo} in ${INSTALL_PREFIX}/${MANDIR}"
	install -c -o root -g ${IDSGRP} -m 444 man/${foo} ${INSTALL_PREFIX}/${MANDIR}/${foo}
done
for foo in ${filters} ; do
	if [ ! -f ${INSTALL_PREFIX}/${ETCDIR}/filters/${foo} ]; then
		echo "Installing etc/filters/${foo} in ${INSTALL_PREFIX}/${ETCDIR}/filters"
		install -c -o root -g ${IDSGRP} -m 440 etc/filters/${foo} ${INSTALL_PREFIX}/${ETCDIR}/filters/${foo}
	fi
done
if [ ! -f ${INSTALL_PREFIX}/${ETCDIR}/sensor.init ]; then
		echo "Installing etc/sensor.init in ${INSTALL_PREFIX}/${ETCDIR}"
		install -c -o root -g ${IDSGRP} -m 550 etc/sensor.init ${INSTALL_PREFIX}/${ETCDIR}/sensor.init
fi

echo "Installing etc/help.txt in ${INSTALL_PREFIX}/${ETCDIR}"
install -c -o root -g ${IDSGRP} -m 440 etc/help.txt ${INSTALL_PREFIX}/${ETCDIR}/help.txt

if [ ! -f /etc/shoki.conf ]; then
	install -c -o root -g ${IDSGRP} -m 440 shoki.conf /etc/shoki.conf
fi

exit 0
