#define ipaddr_string(p) getname((const u_char *)(p))

#define CALLBACK_NULL		0
#define CALLBACK_LOG		1
#define CALLBACK_MAIL		2
#define CALLBACK_SEARCH		3
#define CALLBACK_VERBOSE_LOG	4
#define	CALLBACK_VERBOSE_MAIL	5
#define CALLBACK_VERBOSE_SEARCH	6
#define CALLBACK_TERSE_LOG	7
#define CALLBACK_TERSE_MAIL	8
#define CALLBACK_TERSE_SEARCH	9
#define CALLBACK_PGSQL_LOG	10
#define CALLBACK_PGSQL_SEARCH	11
#define CALLBACK_PGSQL_MAIL	12
#define CALLBACK_REMOTE		13
#define CALLBACK_MESHUGA_TSEQ	14
#define CALLBACK_MESHUGA_T1	15
#define CALLBACK_MESHUGA_T2	16
#define CALLBACK_MESHUGA_T3	17
#define CALLBACK_MESHUGA_T4	18
#define CALLBACK_MESHUGA_T5	19
#define CALLBACK_MESHUGA_T6	20
#define CALLBACK_MESHUGA_T7	21
#define CALLBACK_MESHUGA_PU	22
#define CALLBACK_THRESHOLD	23
#define CALLBACK_HOST_SCAN	24
#define CALLBACK_PORT_SCAN	25

#define PRINT_PACKET	0
#define PRINT_TERSE	1
#define PRINT_POSTGRES	2

#define OUT_BUFSIZE	131072

struct filter
{
	char filter[MAX_FILTER_LEN];            /* filter expression */
	char hex[MAX_HEX_LEN];                  /* search expression */
	char call_args[MAX_CALLARG_LEN];        /* args */
	char id[MAX_ID_LEN];                    /* unique filter id */
	char type[MAX_TYPE_LEN];                /* filter type */
	regex_t *regex;                         /* compiled regex */
	struct bpf_program *fcode;              /* compiled fcode */
	int fcount;                             /* snaplen */
	int d;                                  /* threshold duration */
	int t;                                  /* threshold value */
	int q;                                  /* threshold `quiet' time */
	int h;                                  /* threshold host count */
	int sev;                                /* filter severity */
	int callback;                           /* callback id */
	int index;				/* callback index */
};

int callback_handler(const u_char *, struct filter *, int, const struct pcap_pkthdr *, int);
int callback_log(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
void callback_mail(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
int callback_search(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
void callback_print_log(const u_char *, struct filter *, int, const struct pcap_pkthdr *, int);
int callback_remote(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
int callback_threshold(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
int callback_host_scan(const u_char *, struct filter *, int, const struct pcap_pkthdr *);
int callback_port_scan(const u_char *, struct filter *, int, const struct pcap_pkthdr *);

void pcap_print_pt(u_char *, const struct pcap_pkthdr *, const u_char *);

char *getname(const u_char *ap);
char *intoa(u_int32_t addr);

#define PF_TYPE_VAR	1
#define PF_TYPE_DELIM	2

#define PF_SRC		1
#define	PF_SPORT	2
#define PF_DST		3
#define PF_DPORT	4
#define PF_PROTO	5
#define PF_SEQ		6
#define	PF_DF		7
#define PF_SYN		8
#define PF_FIN		9
#define PF_RST		10
#define PF_PUSH		11
#define PF_TIME		12
#define PF_ICMP_TYPE	13
#define PF_ACK		14
#define PF_LEN		15
#define PF_FLAGS	16
#define PF_NAME		17
#define PF_ID		18
#define PF_DATA		19
#define PF_HEX		20
#define PF_SEC		21
#define PF_USEC		22
#define PF_VER		23
#define PF_TOS		24
#define PF_FRAG		25
#define PF_TTL		26
#define PF_IPSUM	27
#define PF_WIN		28
#define PF_URG		29
#define PF_IPID		30
#define PF_ALERT	31
#define PF_SENSOR	32
#define PF_RUNTIME	33

#define DELIMETER ','

struct dstlist
{
	u_int32_t dst;
	int count;
	struct dstlist *next;
};

struct plist
{
	int port;
	int total;
	struct dstlist *dstlist;
	struct plist *next;
};

struct srclist
{
	u_int32_t src;
	int total;
	struct dstlist *dstlist;
	struct plist *ports;
	struct srclist *next;
};

struct frags
{
	struct pcap_pkthdr *h;
	u_char *p;
	u_short off;
	int len;
	struct frags *next;
	struct frags *prev;
};

struct flist
{
	u_int32_t src;
	u_int32_t dst;
	u_short id;
	u_int8_t tos;
	u_int8_t p;
	int done;
	struct frags *frags;
	struct flist *next;
	struct flist *prev;
};

char *packet_output(const u_char *, const struct pcap_pkthdr *, int format[], char *, char *, const u_char *);
void packet_print(const u_char *, u_int, const u_char *, u_char, const struct pcap_pkthdr *, struct filter *, const u_char *);
void terse_print(const u_char *, u_int, const u_char *, u_char, const struct timeval *, int);
void postgres_print(const u_char *, u_int, const u_char *, u_char, const struct timeval *, int);
int get_callback(char *);
int hs_add(u_int32_t daddr, u_int32_t saddr, int i, int port, int count);
int permute_hosts(u_int32_t saddr, const u_char *sp, const struct pcap_pkthdr *, int outformat[], struct filter *, int t1, int t2);
void log_packet(const u_char *, const struct pcap_pkthdr *, int outformat[], char *, char *, const u_char *);
void log_msg(const struct pcap_pkthdr *, const u_char *sp, const u_char *rsp, char *, ...);
int check_frag(const struct pcap_pkthdr *, const u_char *);
int frag_reassemble(u_short);
u_short ip_checksum(u_char *, int);
void make_filter(int, struct filter *);
void doctrine_to_filter(struct doctrine *, struct filter *);
