#include <sys/ioctl.h>
#include <math.h>

#ifndef __dead
#define __dead volatile
#endif

#ifndef MAX_TCPOPTLEN
#define MAX_TCPOPTLEN	40
#endif

#define COMPILING_MESHUGA
#define SHOKI_TYPE	"MESHUGA"
#define SHOKI_LOG_NAME	SHOKI_LOG_NAME_MESHUGA

#define P_SEQ_TR	1
#define P_SEQ_64K	2
#define P_SEQ_I800	3
#define P_SEQ_C		4
#define P_SEQ_RI	5
#define P_SEQ_TD	6

#define FOPT_NAME	1
#define FOPT_TSEQ	2
#define FOPT_T1		3
#define FOPT_T2		4
#define FOPT_T3		5
#define FOPT_T4		6
#define FOPT_T5		7
#define FOPT_T6		8
#define FOPT_T7		9
#define FOPT_PU		10

#define FTOPT_RESP	1
#define FTOPT_DF	2
#define FTOPT_W		3
#define FTOPT_ACK	4
#define FTOPT_FLAGS	5
#define FTOPT_OPS	6

#define FSOPT_CLASS	1
#define FSOPT_GCD	2
#define FSOPT_SI	3
#define FSOPT_VAL	4

int cleanup(int);
void change_log();
void open_dump();
void capture_loop();
void close_dump();
void change_dump();
void clean_kids();

void compile_filters(pcap_t *, bpf_u_int32, int);
void compile_fcodes(pcap_t *, bpf_u_int32, int);
void compile_regexen();
void parse_sensor_config(const char *fname);
void pcap_dump_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp);

int sendpacket(int, u_char *, int);
unsigned short ip_checksum(unsigned char *, int);
void callback_meshuga(const u_char *, int, int, const struct pcap_pkthdr *);
int get_bpf(char *);
void parse_shoki_config(const char *);
void default_formats(int d_f[], char *d_d[], int d_t[]);
char *get_filter(char **);

void read_nmap_fingerprints();
u_int32_t seq_tr();
u_int32_t seq_64k(u_int32_t);
u_int32_t seq_i800(u_int32_t);
u_int32_t seq_c(u_int32_t, u_int32_t);
u_int32_t seq_ri(u_int32_t, int, u_int32_t);
u_int32_t seq_td(u_int32_t, int, u_int32_t);
struct fttest *parse_fttest(char *);
struct futest *parse_futest(char *);
struct ftseq *parse_ftseq(char *);

u_int32_t get_ack(char *type, u_int32_t current);

#define ipaddr_string(p) getname((const u_char *)(p))
#define BIGMOD (u_int32_t)(pow(2, (sizeof(u_int32_t) * 8)) - 1)


struct
pseudo_header
{
        unsigned long s_addr;
        unsigned long d_addr;
        char zero;
        unsigned char proto;
        unsigned short length;
};

struct
fttest
{
        int resp;
        int df;
        int w;
        char *ack;
        u_char flags;
        u_char tcp_ops[MAX_TCPOPTLEN];
        int tcp_oplen;
};

struct
futest
{
        int resp;
        int df;
        int tos;
        int iplen;
        int riptl;
        int rid;
        int ripck;
        int uck;
        int ulen;
        char *dat;
};

struct
ftseq
{
        int class;
        int gcd;
        int si;
        int val;
};

struct
fprint
{
        char *name;
        struct ftseq *ftseq;
        struct fttest *t1;
        struct fttest *t2;
        struct fttest *t3;
        struct fttest *t4;
        struct fttest *t5;
        struct fttest *t6;
        struct fttest *t7;
        struct futest *pu;
};

struct
fprint_fake
{
        char *name;
        struct tseq *tseq;
        char *t1;
        char *t2;
        char *t3;
        char *t4;
        char *t5;
        char *t6;
        char *t7;
        char *pu;
};

