#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_meshuga.h"
#include "shoki_config.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"

extern char *fprint_file;
extern int nfilters;
int debug = 0, sent = 0, echo_mss = 0, bpf_fd, fragre = 0, open_ports[65535];
int silent = 0, callbacks = 1, timeval_kludge = 0, linux_kludge = 0;
struct bpf_program base_fcode;
pcap_handler out = pcap_dump_pt;
u_char *pcap_userdata = 0;
struct fprint *fprint;
u_int32_t old_seq = 0;
FILE *logfile = NULL;

int snaplen = DEFAULT_SNAPLEN;

static struct { const char *name; int fopt; } fopts[] = 
{
	{ "NULL", 0 },
	{ "fingerprint", 1 },
	{ "tseq", 2 },
	{ "t1", 3 },
	{ "t2", 4 },
	{ "t3", 5 },
	{ "t4", 6 },
	{ "t5", 7 },
	{ "t6", 8 },
	{ "t7", 9 },
	{ "pu", 10 }
};

static struct { const char *name; int op; } fttestop[] =
{
	{ "NULL", 0 },
	{ "resp", 1 },
	{ "df", 2 },
	{ "w", 3 },
	{ "ack", 4 },
	{ "flags", 5 },
	{ "ops", 6 }
};

static struct { const char *name; int op; } ftseqop[] =
{
	{ "NULL", 0 },
	{ "class", 1 },
	{ "gcd", 2 },
	{ "si", 3 },
	{ "val", 4 }
};

#define COMMA ","

int main(int argc, char **argv)
{
	int op;
	char *cfile = NULL;
	bpf_u_int32 localnet, netmask;
	char errbuf[PCAP_ERRBUF_SIZE];
	struct sigaction *cleansig;
	struct sigaction *childsig;
	int dflag = 0;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "c:C:dF:i:s:v")) != EOF)
		switch(op)
		{
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'd':
				dflag++;
				break;
			case 'F':
				fname = optarg;
				break;
			case 'i':
				dev = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'v':
				debug++;
				break;
		}

	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(MESHUGA_CONFIG)) : 
		(parse_shoki_config(cfile)));
	(void)read_nmap_fingerprints(fprint_file);
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(dev == NULL) error("Specify a device, dammit.");
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "meshuga starting");
       	(void)closelog();
	if(!(dflag)) if(fork()) exit(0);
	pd = pcap_open_live(dev, snaplen, 1, 1000, errbuf);
	if(pd == NULL) error("%s", errbuf);
	localnet = 0;
	netmask = 0;
	if(pcap_lookupnet(dev, &localnet, &netmask, errbuf) < 0)
		((localnet = 0), (netmask = 0), error("%s", errbuf));
	if(pcap_compile(pd, &base_fcode, base_filter, 1, netmask) < 0) 
		error("compiling base_fcode: %s\n", pcap_geterr(pd));
	if(pcap_setfilter(pd, &base_fcode) < 0) 
		error("setting base_fcode:  %s\n", pcap_geterr(pd));
	if(nfilters > 0) compile_filters(pd, netmask, 0);
	bpf_fd = get_bpf(dev);
	portlist = strtok(portlist, COMMA);
	open_ports[atoi(portlist)] = 1;
	while((portlist = strtok(NULL, COMMA)))
		open_ports[atoi(portlist)] = 1;
	for(;;) (void)capture_loop();
	pcap_close(pd);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL) (void)fflush(stdout);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "meshuga shutting down");
	exit(0);
}

void
capture_loop()
{
	if(pcap_loop(pd, count, out, pcap_userdata) < 0)
		error("%s", pcap_geterr(pd));
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_dump_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	struct pcap_pkthdr *h_pt;
	int i;

	raise(SIGCHLD);
	sent = 0;
	for(i = 1; i <= nfilters; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			h_pt = (struct pcap_pkthdr *)(char *)h;
			h_pt->caplen = ((fcount_pt[i] < h->caplen) ? fcount_pt[i] : h->caplen);
			if(callback_pt[i] == 0) return;
			callback_meshuga((u_char *)sp, i, h_pt->caplen, h);
		}
	}
}

void
callback_meshuga(const u_char *sp, int i, int len, const struct pcap_pkthdr *h)
{
	struct ether_header *ep, *p_ep;
	struct ip *ip, *p_ip;
	struct tcphdr *tp, *p_tp;
	u_char *p_ip_opts;
	u_char *p_tcp_opts = NULL, *p_tp_opts = NULL;
	u_char *packet;
	u_short s_sport, s_dport, s_win;
	u_short p_win = 0;
	u_int32_t s_seq, s_ack;
	u_int32_t p_seq = 0, p_ack = 0;
	struct pseudo_header *pseudo;
	u_char s_flags, p_flags = 0;
	int s_hl;
	int p_v = 4, p_hl = 5, p_tos = 0, p_off = 0, p_ttl = 64;
	int p_tp_oplen = 0, p_oplen;
	int p_p = IPPROTO_TCP;
	int sendit = 0;
	u_int length = h->len;
	u_short ether_type;
	extern int bpf_fd;
	int mss, j, k = 0;
	char *baz;
	u_char *tp_opts = NULL;


	length = sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct tcphdr) + 40 + 40;
	if(!(ep = (struct ether_header *)calloc(1, sizeof(struct ether_header))))
		error("calloc:  callback_meshuga");
	memcpy(ep, sp, sizeof(struct ether_header));
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP) return;
	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  callback_meshuga");
	memcpy(ip, (sp + sizeof(struct ether_header)), sizeof(struct ip));
	if(ip->ip_p != IPPROTO_TCP) return;
	s_hl = ip->ip_hl;
	if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  callback_meshuga");
	memcpy(tp, (sp + sizeof(struct ether_header) + sizeof(struct ip) + ((s_hl * 4) - 20)), sizeof(struct tcphdr));
	p_oplen = (tp->th_off << 2) - sizeof(struct tcphdr);
	if(p_oplen > 0)
	{
		if(!(tp_opts = (u_char *)calloc(1, p_oplen)))
			error("calloc:  callback_meshuga");
		memcpy(tp_opts, (sp + sizeof(struct ether_header) + sizeof(struct ip) + ((s_hl * 4) - 20) + sizeof(struct tcphdr)), p_oplen);
	}
	s_sport = ntohs(tp->th_sport);
	s_dport = ntohs(tp->th_dport);
	s_seq = ntohl(tp->th_seq);
	s_ack = ntohl(tp->th_ack);
	s_win = ntohs(tp->th_win);
	s_flags = tp->th_flags;

	if(!(packet = (u_char *)calloc(1, length)))
		error("calloc:  callback_meshuga");
	p_ep = (struct ether_header *)packet;
	p_ip = (struct ip *)((char *)packet + (sizeof(struct ether_header)));
	p_ip_opts = (char *)((char *)packet + (sizeof(struct ether_header)) + (sizeof(struct ip)));

	switch(callback_pt[i])
	{
		case CALLBACK_MESHUGA_TSEQ:
			if(sent) break;
			if(open_ports[s_dport] == 0)
			{
				p_ack = s_seq + 1;
				p_flags = p_flags | TH_RST;
			}
			else
			{
				p_ack = s_seq + 1;
				p_flags = p_flags | TH_SYN;
				p_flags = p_flags | TH_ACK;
			}
			sendit = 9;
			break;
		case CALLBACK_MESHUGA_T1:
			if(sent) break;
			if(open_ports[s_dport] == 0) break;
			if(fprint->t1->resp != 0) sendit = 1;
			p_win = fprint->t1->w;
			p_ack = get_ack(fprint->t1->ack, s_seq);
			p_flags = fprint->t1->flags;
			p_tcp_opts = (u_char *)fprint->t1->tcp_ops;
			p_tp_oplen = fprint->t1->tcp_oplen;
			if(fprint->t1->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T2:
			if(sent) break;
			if(open_ports[s_dport] == 0) break;
			if(fprint->t2->resp != 0) sendit = 2;
			p_win = fprint->t2->w;
			p_ack = get_ack(fprint->t2->ack, s_seq);
			p_flags = fprint->t2->flags;
			p_tcp_opts = (u_char *)fprint->t2->tcp_ops;
			p_tp_oplen = fprint->t2->tcp_oplen;
			if(fprint->t2->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T3:
			if(sent) break;
			if(open_ports[s_dport] == 0) break;
			p_ack = get_ack(fprint->t3->ack, s_seq);
			if(fprint->t3->resp != 0) sendit = 3;
			p_win = fprint->t3->w;
			p_flags = fprint->t3->flags;
			p_tcp_opts = (u_char *)fprint->t3->tcp_ops;
			p_tp_oplen = fprint->t3->tcp_oplen;
			if(fprint->t3->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T4:
			if(sent) break;
			if(open_ports[s_dport] == 0) break;
			p_ack = get_ack(fprint->t4->ack, s_seq);
			if(fprint->t4->resp != 0) sendit = 4;
			p_win = fprint->t4->w;
			p_flags = fprint->t4->flags;
			p_tcp_opts = (u_char *)fprint->t4->tcp_ops;
			p_tp_oplen = fprint->t4->tcp_oplen;
			if(fprint->t4->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T5:
			if(sent) break;
			if(open_ports[s_dport] == 1) break;
			p_ack = get_ack(fprint->t5->ack, s_seq);
			if(fprint->t5->resp != 0) sendit = 5;
			p_win = fprint->t5->w;
			p_flags = fprint->t5->flags;
			p_tcp_opts = (u_char *)fprint->t5->tcp_ops;
			p_tp_oplen = fprint->t5->tcp_oplen;
			if(fprint->t5->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T6:
			if(sent) break;
			if(open_ports[s_dport] == 1) break;
			p_ack = get_ack(fprint->t6->ack, s_seq);
			if(fprint->t6->resp != 0) sendit = 6;
			p_win = fprint->t6->w;
			p_flags = fprint->t6->flags;
			p_tcp_opts = (u_char *)fprint->t6->tcp_ops;
			p_tp_oplen = fprint->t6->tcp_oplen;
			if(fprint->t6->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_T7:
			if(sent) break;
			if(open_ports[s_dport] == 1) break;
			p_ack = s_seq + 1;
			if(fprint->t7->resp != 0) sendit = 7;
			p_ack = get_ack(fprint->t7->ack, s_seq);
			p_win = fprint->t7->w;
			p_flags = fprint->t7->flags;
			p_tcp_opts = (u_char *)fprint->t7->tcp_ops;
			p_tp_oplen = fprint->t7->tcp_oplen;
			if(fprint->t7->df == 1) p_off |= IP_DF;
			break;
		case CALLBACK_MESHUGA_PU:
			if(open_ports[s_dport] == 1) break;
			p_flags = p_flags | IP_DF;
			break;
	}

	if(sendit == 0)
	{
		free(tp);
		free(ip);
		free(ep);
		free(packet);
		return;
	}
	
	memcpy(p_ep->ether_dhost, ep->ether_shost, sizeof("FF:FF:FF:FF:FF:FF"));
	memcpy(p_ep->ether_shost, ep->ether_dhost, sizeof("FF:FF:FF:FF:FF:FF"));
	p_ep->ether_type = htons(ETHERTYPE_IP);

	if(old_seq == 0) old_seq = seq_tr();
	switch(fprint->ftseq->class)
	{
		case P_SEQ_64K:
			p_seq = seq_64k(old_seq);
			break;
		case P_SEQ_RI:
			p_seq = seq_ri(old_seq, fprint->ftseq->gcd, fprint->ftseq->si);
			break;
		case P_SEQ_C:
			p_seq = seq_c(old_seq, fprint->ftseq->val);
			break;
		case P_SEQ_TR:
			p_seq = seq_tr();
			break;
		case P_SEQ_I800:
			p_seq = seq_i800(old_seq);
			break;
		case P_SEQ_TD:
			p_seq = seq_td(old_seq, fprint->ftseq->gcd, fprint->ftseq->si);
			break;
	}

	p_tp = (struct tcphdr *)((char *)packet + sizeof(struct ether_header) + sizeof(struct ip) + ((p_hl * 4) - 20));
	p_tp_opts = (u_char *)((char *)packet + sizeof(struct ether_header) + sizeof(struct ip) + ((p_hl * 4) - 20) + sizeof(struct tcphdr));
	memcpy(p_tp_opts, p_tcp_opts, p_tp_oplen);
	baz = (u_char *)((char *)packet + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct tcphdr) + ((p_hl * 4) - 20));
	if((echo_mss != 0) && (p_oplen > 0))
	{
		k = 0;
		for(j = 0; (j <= p_oplen) && (k == 0) ; j++)
		{
			switch(tp_opts[j])
			{
				case TCPOPT_MAXSEG:
					memcpy(&mss, &tp_opts[j + 2], 2);
					k++;
					break;
				case TCPOPT_EOL:
					k++;
					break;
				default:
					j += tp_opts[j + 1];
					break;
			}
		};
		if(k == 0) mss = htons(265);
		k = 0;
		for(j = 0; (j <= p_tp_oplen) && (k == 0); j++)
		{
			switch(baz[j])
			{
				case TCPOPT_MAXSEG:
					memcpy(&baz[j + 2], &mss, 2);
					k++;
					break;
				case TCPOPT_EOL:
					k++;
					break;
				default:
					j += baz[j + 1];
					break;
			}
		}
	}
	p_tp->th_sport = tp->th_dport;
	p_tp->th_dport = tp->th_sport;
	p_tp->th_ack = htonl(p_ack);
	p_tp->th_seq = htonl(p_seq % BIGMOD);
	p_tp->th_off = (sizeof(struct tcphdr) + p_tp_oplen) >> 2;
	p_tp->th_flags = p_flags;
	p_tp->th_win = htons(p_win);
	p_tp->th_sum = 0;
	p_tp->th_urp = tp->th_urp;
	old_seq = p_seq;

	p_ip->ip_v = p_v;
	p_ip->ip_hl = p_hl;
	p_ip->ip_tos = p_tos;
	p_ip->ip_id = htons(666);
	p_ip->ip_off = htons(p_off);
	p_ip->ip_len = htons(sizeof(struct ip) + sizeof(struct tcphdr) + ((p_hl *4) - 20) + p_tp_oplen);
	p_ip->ip_ttl = p_ttl;
	p_ip->ip_p = p_p;
	p_ip->ip_sum = 0;
	p_ip->ip_src.s_addr = ip->ip_dst.s_addr;
	p_ip->ip_dst.s_addr = ip->ip_src.s_addr;

	p_ip->ip_sum = ip_checksum((unsigned char *)p_ip, sizeof(struct ip));

	/* Drunk.  Fix later. -spb */
	if(!(pseudo = (struct pseudo_header *)calloc(1, (sizeof(struct pseudo_header) + sizeof(struct tcphdr)))))
		error("calloc:  callback_meshuga");
	pseudo->s_addr = p_ip->ip_src.s_addr;
	pseudo->d_addr = p_ip->ip_dst.s_addr;
	pseudo->proto = IPPROTO_TCP;
	pseudo->length = htons(sizeof(struct tcphdr));
	p_tp->th_sum = ip_checksum((unsigned char *)pseudo, sizeof(struct pseudo_header));
	p_tp->th_sum = ~p_tp->th_sum;
	p_tp->th_sum = ip_checksum((unsigned char *)p_tp, sizeof(struct tcphdr));
	free(pseudo);

	if(debug)
	{
		(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       		syslog(SHOKI_LOG_LVL, "T%d probe from %s, port %d", sendit, ipaddr_string(&p_ip->ip_dst), s_dport);
       		(void)closelog();
	}
	(void)sendpacket(bpf_fd, packet, length);
	sent++;
	if(p_oplen > 0) free(tp_opts);
	free(tp);
	free(ip);
	free(ep);
	free(packet);
}

int
sendpacket(int fd, u_char *packet, int length)
{
	if(write(fd, packet, length) == -1)
		error("sendpacket error");
	return(length);
}

#ifndef NBPFILTER
#define NBPFILTER 32
#endif

int
get_bpf(char *dev)
{
	struct bpf_version ver;
	struct ifreq ifr;
	struct timeval tv;
	char bpfdev[NBPFILTER];
	int fd = 0, i;

	for(i = 0; i < NBPFILTER; i++)
	{
		(void)sprintf(bpfdev, "/dev/bpf%d", i);
		if((fd = open(bpfdev, O_RDWR)) >= 0) break;
	}
	if(i == NBPFILTER) error("Checked %d bpf devices; none available", i);
	if(ioctl(fd, BIOCVERSION, (caddr_t)&ver) == -1) error("BIOCVERSION");
	if(ver.bv_major != BPF_MAJOR_VERSION ||
		ver.bv_minor < BPF_MINOR_VERSION)
		error("BPF_[MAJOR|MINOR]_VERSION");
	(void)strncpy(ifr.ifr_name, dev, sizeof(ifr.ifr_name));
	if(ioctl(fd, BIOCSETIF, &ifr) == -1) error("BIOCSETIF");
	if(ioctl(fd, BIOCPROMISC, NULL) == -1) error("BIOCPROMISC");
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if(ioctl(fd, BIOCSRTIMEOUT, (caddr_t)&tv) == -1)
		error("BIOCSRTIMEOUT");
	(void)ioctl(fd, BIOCFLUSH, 0);
	return(fd);
}

/* TR my ass.  Don't use any of this for real isn generation.  -spb*/
u_int32_t
seq_tr()
{
	u_int32_t foo;
	u_int bar;
	bar = (u_int)(1 + (u_int)(BIGMOD * (random()/(RAND_MAX + 1.0))) % 65699);
	srandom(bar);
	foo = 1 + (int)((random() * BIGMOD)/(RAND_MAX + 1.0));
	return(foo);
}

u_int32_t
seq_64k(u_int32_t seq)
{
	u_int32_t foo;
	foo = seq + 64000;
	return(foo);
}

u_int32_t
seq_i800(u_int32_t seq)
{
	u_int32_t foo;
	foo = seq + 800;
	return(foo);
}

u_int32_t
seq_c(u_int32_t seq, u_int32_t bar)
{
	u_int32_t foo;
	foo = seq + bar;
	return(foo);
}

u_int32_t
seq_ri(u_int32_t seq, int gcd, u_int32_t si)
{
	u_int32_t foo;
	if(gcd == 0) gcd = 1 + (int)(1024 * (random()/(RAND_MAX + 1.0)));
	if(si == 0) si = 1 + (int)(1024 * (random()/(RAND_MAX + 1.0)));
	foo = seq + (gcd *  si);
	return(foo);
}

u_int32_t
seq_td(u_int32_t seq, int gcd, u_int32_t si)
{
	u_int32_t foo;
	if(gcd == 0) gcd = 1 + (int)(1024 * (random()/(RAND_MAX + 1.0)));
	if(si == 0) si = 1 + (int)(1024 * (random()/(RAND_MAX + 1.0)));
	foo = (u_int32_t)(seq + (gcd * (si * si)));
	return(foo);
}

#define WHITE " \t("
void
read_nmap_fingerprints(const char *config)
{
	FILE *f = NULL;
	char line[1024];
	char *foo, *bar;
	struct fprint_fake *fprint_fake;
	int i, j, match = -1, lno = 0;

	if(!(fprint = (struct fprint *)calloc(1, sizeof(struct fprint))))
		error("calloc:  read_nmap_fingerprints");
	if(!(fprint_fake = (struct fprint_fake *)calloc(1, sizeof(struct fprint_fake))))
		error("calloc:  read_nmap_fingerprints");
	f = fopen(config, "r");
	if(f == NULL) error("Couldn't read %s", config);
	while(fgets(line, sizeof(line), f))
	{
		lno++;
		foo = line + strspn(line, WHITE);
		if(*foo == '#' || *foo == '\n' || !*foo) continue;
		foo = strtok(foo, WHITE);
		for(i=0; foo[i]; i++) foo[i] = tolower(foo[i]);
		match = -1;
		for(j=0; fopts[j].name && match == -1; j++)
		{
			if(strcmp(foo, fopts[j].name) == 0)
				match = fopts[j].fopt;
		}
		if(match == -1) error("fingerprint parsing error, line (%d) (%s)", lno, foo);
		foo = line + strlen(foo) + 1;
		while(*foo && isspace((int)*foo)) foo++;
		bar = foo;
		while(*bar && *bar != '\n' && *bar != '#') bar++;
		while(isspace((int)*bar)) bar--;
		bar++;
		*bar = '\0';
		if(!*foo) error("No argument");
		switch(match)
		{
			case FOPT_NAME:
				if(fprint->name != NULL && !(memcmp(fprint->name, ostype, strlen(ostype))))
					goto done;
				fprint->name = strdup(foo);
				break;
			case FOPT_TSEQ:
				fprint->ftseq = (struct ftseq *)parse_ftseq(foo);
				break;
			case FOPT_T1:
				fprint_fake->t1 = strdup(foo);
				break;
			case FOPT_T2:
				fprint_fake->t2 = strdup(foo);
				break;
			case FOPT_T3:
				fprint_fake->t3 = strdup(foo);
				break;
			case FOPT_T4:
				fprint_fake->t4 = strdup(foo);
				break;
			case FOPT_T5:
				fprint_fake->t5 = strdup(foo);
				break;
			case FOPT_T6:
				fprint_fake->t6 = strdup(foo);
				break;
			case FOPT_T7:
				fprint_fake->t7 = strdup(foo);
				break;
			case FOPT_PU:
				fprint_fake->pu = strdup(foo);
				break;
		}
	}
	error("Fingerprint not found");
	return;
done:
	fprint->t1 = (struct fttest *)parse_fttest(fprint_fake->t1);
	fprint->t2 = (struct fttest *)parse_fttest(fprint_fake->t2);
	fprint->t3 = (struct fttest *)parse_fttest(fprint_fake->t3);
	fprint->t4 = (struct fttest *)parse_fttest(fprint_fake->t4);
	fprint->t5 = (struct fttest *)parse_fttest(fprint_fake->t5);
	fprint->t6 = (struct fttest *)parse_fttest(fprint_fake->t6);
	fprint->t7 = (struct fttest *)parse_fttest(fprint_fake->t7);
	fprint->pu = (struct futest *)parse_futest(fprint_fake->pu);
}

#define FDELIM	"%"
#define FWHITE	" =\t"

struct fttest *
parse_fttest(char *line)
{
	struct fttest *test;
	char *foo, *bar, *baz = NULL;
	int i, j, match, mss;

	if(!(test = (struct fttest *)calloc(1, sizeof(struct fttest))))
		error("calloc:  parse_fttest");
	test->resp = 1;
	foo = line + strspn(line, FWHITE);
	if(*foo == '#' || !*foo) return(test);
	while(foo != NULL)
	{
		bar = strdup(foo);
		bar = strtok(bar, FDELIM);
		bar = strtok(bar, FWHITE);
		for(i = 0; bar[i]; i++) bar[i] = tolower(bar[i]);
		match = -1;
		for(j = 0; fttestop[j].name && match == -1; j++)
		{
			if(strcmp(bar, fttestop[j].name) == 0)
				match = fttestop[j].op;
		}
		if(match == -1) error("Bad tcp test option");
		bar = strtok(NULL, FWHITE);
		switch(match)
		{
			case FTOPT_RESP:
				if(strncasecmp(bar, "y", 1) == 0)
					test->resp = 2;
				if(strncasecmp(bar, "n", 1) == 0)
					test->resp = 0;
				break;
			case FTOPT_DF:
				if(strncasecmp(bar, "y", 1) == 0)
					test->df = 1;
				break;
			case FTOPT_W:
				test->w = strtol(bar, NULL, 16);
				break;
			case FTOPT_ACK:
				test->ack = strdup(bar);
				break;
			case FTOPT_FLAGS:
				if(strstr(bar, "F") != NULL)
					test->flags = test->flags | TH_FIN;
				if(strstr(bar, "S") != NULL)
					test->flags = test->flags | TH_SYN;
				if(strstr(bar, "R") != NULL)
					test->flags = test->flags | TH_RST;
				if(strstr(bar, "P") != NULL)
					test->flags = test->flags | TH_PUSH;
				if(strstr(bar, "A") != NULL)
					test->flags = test->flags | TH_ACK;
				if(strstr(bar, "U") != NULL)
					test->flags = test->flags | TH_URG;
				break;
			case FTOPT_OPS:
				baz = strdup(bar);
				while(*baz != NULL)
				{
					if(*baz == 'L')
					{
						test->tcp_ops[test->tcp_oplen] = TCPOPT_EOL;
						test->tcp_oplen++;
					}
					if(*baz == 'N')
					{
						test->tcp_ops[test->tcp_oplen] = TCPOPT_NOP;
						test->tcp_oplen++;
					}
					if(*baz == 'M')
					{
						test->tcp_ops[test->tcp_oplen] = TCPOPT_MAXSEG;
						test->tcp_ops[test->tcp_oplen + 1] = TCPOLEN_MAXSEG;
						if(strstr(bar, "E") != NULL)
						{
							/* Crude */
							echo_mss = 1;
						}
						else
						{
							/* Arbitrary */
							mss = htons(512);
							memcpy(&test->tcp_ops[test->tcp_oplen + 2], &mss, 2);
						}
						test->tcp_oplen += 4;
					}
					if(*baz == 'W')
					{
						test->tcp_ops[test->tcp_oplen] = TCPOPT_WINDOW;
						test->tcp_ops[test->tcp_oplen + 1] = TCPOLEN_WINDOW;
						test->tcp_ops[test->tcp_oplen + 2 ] = htons(1);
						test->tcp_oplen += 3;
					}
					if(*baz == 'T')
					{
						test->tcp_ops[test->tcp_oplen] = TCPOPT_TIMESTAMP;
						test->tcp_ops[test->tcp_oplen + 1] = TCPOLEN_TIMESTAMP;
						test->tcp_oplen += 10;
					}
					baz++;
				}
				break;
		}
		if((foo = strstr(foo, FDELIM)) == NULL) continue;
		foo++;
	}
	return(test);
}

struct futest *
parse_futest(char *line)
{
	struct futest *test;
	if(!(test = (struct futest *)calloc(1, sizeof(struct futest))))
		error("calloc:  parse_futest");
	return(test);
}

struct ftseq *
parse_ftseq(char *line)
{
	char *foo, *bar;
	struct ftseq *seq;
	int i, j, match = -1, mod = 0;

	if(!(seq = (struct ftseq *)calloc(1, sizeof(struct ftseq))))
		error("calloc:  parse_ftseq");
	foo = line + strspn(line, FWHITE);
	if(*foo == '#' || !*foo) return(0);
	while(foo != NULL)
	{
		bar = strdup(foo);
		bar = strtok(bar, FDELIM);
		bar = strtok(bar, FWHITE);
		for(i = 0; bar[i]; i++) bar[i] = tolower(bar[i]);
		match = -1;
		for(j = 0; ftseqop[j].name && match == -1; j++)
		{
			if(strcmp(bar, ftseqop[j].name) == 0)
				match = ftseqop[j].op;
		}
		if(match == -1) error("Bad tcp seq option");
		bar = strtok(NULL, FWHITE);
		for(i = 0; bar[i]; i++) bar[i] = tolower(bar[i]);
		switch(match)
		{
			case FSOPT_CLASS:
				if(strstr(bar, "tr") != NULL)
					seq->class = P_SEQ_TR;
				if(strstr(bar, "64k") != NULL)
					seq->class = P_SEQ_64K;
				if(strstr(bar, "i800") != NULL)
					seq->class = P_SEQ_I800;
				if(strstr(bar, "c") != NULL)
					seq->class = P_SEQ_C;
				if(strstr(bar, "ri") != NULL)
					seq->class = P_SEQ_RI;
				if(strstr(bar, "td") != NULL)
					seq->class = P_SEQ_TD;
				break;
			case FSOPT_GCD:
				mod = 0;
				if((bar = strtok(bar, "<"))) mod = 1;
				if((bar = strtok(bar, ">"))) mod = -1;
				seq->gcd = (strtol(bar, NULL, 16) + mod);
				break;
			case FSOPT_SI:
				seq->si = strtol(bar, NULL, 16);
				break;
			case FSOPT_VAL:
				seq->val = atoi(bar);
				break;
		}
		if((foo = strstr(foo, FDELIM)) == NULL) continue;
		foo++;
	}
	return(seq);
}

void
pcap_print_pt(u_char *foo, const struct pcap_pkthdr *bar, const u_char *baz)
{
}

u_int32_t
get_ack(char *type, u_int32_t current)
{
	if(type == NULL) return(current + 1);
	if(!(strncasecmp(type, "s++", 3))) return(current + 1);
	if(!(strncasecmp(type, "s", 1))) return(current);
	if(!(strncasecmp(type, "0", 1))) return(0);
	if(!(strncasecmp(type, "O", 1))) return(0);
	return(current + 1);
}
