#!/bin/sh
#
# This is just a little script used to initialise a postgres database for
# keeping track of incidents.

# Comment out the following two lines and set DIR below
#
#
#echo "Read this script before running it."
#exit 1

# /tmp is probably a bad choice.  If you don't understand about /tmp
# race conditions, set this to something else.
DIR=/tmp
#
# Set the value of DB to whatever you want the database to be called.
# Make sure you don't already have a database of the same name.
DB=sensors
#
# Set the value of TABLE to whatever you want the table to be called.
TABLE=sensors
#
# Set the value of LUSER to whatever luser you're planning on running
# the data collection/parsing/analysis widgetry as.  Note that this
# is the luser on the central machine(s) that's doing the work---not
# the sensor.
LUSER=root

cat <<EOF > ${DIR}/incidents.psql
create table ${TABLE}
	(timestamp int4,
	dotsec int4,
	src varchar(15),
	sport int,
	dst varchar(15),
	dport int,
	proto int,
	len int,
	id int,
	seq int,
	ack int,
	tos int,
	ttl int,
	ipid int,
	win int,
	icmp_type int,
	flags varchar(8),
	comment varchar(128),
	alert varchar(128),
	domain varchar(128),
	whois varchar(640),
	trace varchar(1024),
	dump varchar(4096),
	hex varchar(4096),
	sensor varchar(128),
	uniqid varchar(128),
	level int4,
	typeflag int4,
	newflag int4,
	reporttime int4);
create function abstime_datetime(int4)
        returns datetime
        as '' language 'internal';
create function datetime(int4)
        returns datetime
        as 'select abstime_datetime(\$1)' language 'sql';
create function reltime_timespan(int4)
        returns timespan
        as '' language 'internal';
create function timespan(int4)
        returns timespan
        as 'select reltime_timespan(\$1)' language 'sql';
create view average_hours_between_incidents
	as select (((max(timestamp) - min(timestamp)) / (60 * 60)) / count(*))
	from ${TABLE};
create view summary
	as SELECT "datetime"("${TABLE}"."timestamp") AS "datetime",
	"${TABLE}"."timestamp", "${TABLE}"."dotsec", "${TABLE}"."src",
	"${TABLE}"."sport", "${TABLE}"."dst", "${TABLE}"."dport",
	"${TABLE}"."proto" 
	from ${TABLE};
create view dateview
	as SELECT "datetime"("${TABLE}"."timestamp") AS "datetime",
	"${TABLE}"."timestamp", "${TABLE}"."dotsec", "${TABLE}"."src",
	"${TABLE}"."sport", "${TABLE}"."dst", "${TABLE}"."dport",
	"${TABLE}"."proto", "${TABLE}"."comment", "${TABLE}"."alert",
	"${TABLE}"."whois", "${TABLE}"."trace", "${TABLE}"."dump"
	from ${TABLE};
create view new_events
	as SELECT "datetime"("${TABLE}"."timestamp") AS "datetime",
	"${TABLE}"."timestamp", "${TABLE}"."dotsec", "${TABLE}"."alert",
	"${TABLE}"."src", "${TABLE}"."sport",
	"${TABLE}"."dst", "${TABLE}"."dport",
	"${TABLE}"."proto"
	from ${TABLE} where ("${TABLE}"."newflag" = 1);
create index saddr on ${TABLE} using btree (src);
create index daddr on ${TABLE} using btree (dst);
create user ${LUSER};
grant select on ${TABLE} to ${LUSER};
grant insert on ${TABLE} to ${LUSER};
grant select on summary to ${LUSER};
grant select on dateview to ${LUSER};
grant select on average_hours_between_incidents to ${LUSER};
grant select on new_events to ${LUSER};
EOF
chmod 444 ${DIR}/incidents.psql

su -c "createdb ${DB}" - postgres
su -c "psql -d ${DB} -f ${DIR}/incidents.psql" - postgres


exit 0
