#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_doctrine.h"
#include "shoki_config.h"
#include "shoki_callback.h"
#include "shoki_parser.h"

u_char *pcap_userdata = 0;
char *dumpfile = NULL, *real_outfile = NULL;
int callbacks = 1, fragre = 0, linux_kludge = 0, timeval_kludge = 0;
int rewrite = 0, silent = 0;
int firstsec, lastsec, malsize;
struct bpf_program base_fcode;
FILE *logfile = NULL;

pcap_handler out = pcap_print_pt;

extern int nfilters;

int main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL;
	register char *config = NULL;
	struct sigaction *cleansig;
	struct sigaction *childsig;
	int Pflag = 0;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "Bc:C:E:fF:kKL:Pr:s:S:t:vw:")) != EOF)
		switch(op)
		{
			case 'B':
				callbacks = 0;
				break;
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'E':
				stop = atoi(optarg);
				lastsec = stop;
				break;
			case 'f':
				fragre = 1;
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'L':
				currentfile = optarg;
				break;
			case 'P':
				Pflag++;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);
				firstsec = start;
				break;
			case 't':
				shoki_type = optarg;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'w':
				real_outfile = optarg;
				rewrite = 1;
				break;
		}

	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(PARSER_CONFIG)) : 
		(parse_shoki_config(cfile)));
	if((pf_outformat[0] == 0) || (pf_outdelim[0] == NULL))
		(void)default_formats(pf_outformat_d, pf_outdelim_d, pf_type_d);
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(filterlist == NULL) error("Specify a filterlist, dammit.");
	config = read_file(filterlist);
	if(dumpfile == NULL)
		error("Specify a dump file, dammit.");
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	pd = fake_pcap(pd, 0);
	if(Pflag != 0) currentfile = NULL;
	/* Bad developer.  Bad.  No biscuit. */
	if(currentfile != NULL)
	{
		if(*currentfile == '-')
			logfile = stdout;
		else
			logfile = fopen(currentfile, "a");
		setlinebuf(logfile);
		fflush(logfile);
	}
       	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "parser starting");
	if(rewrite) (void)open_dump();
	(void)read_zdumps(dumpfile, out);
	if(logfile) (void)fclose(logfile);
	cleanup(0);
	if(dumper) (void)pcap_dump_close(dumper);
	exit(0);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL)
	{
		(void)fflush(stdout);
		putc('\n', stderr);
	}
       	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "parser shutting down");
	if(dumper) (void)pcap_dump_close(dumper);
	exit(0);
}

void
open_dump()
{
	int i;
	if(start && stop)
	{
		malsize = strlen(outfile) + (sizeof("123456780") * 2) + 3;
		if(!(currentdump = calloc(1, malsize)))
			error("calloc:  open_dump");
		(void)snprintf(currentdump, malsize, "%s.%010d-%010d", outfile, start, stop);
	}
	else
	{
		i = time(NULL);
		malsize = strlen(outfile) + sizeof("123456780") + 3;
		if(!(currentdump = calloc(1, malsize)))
			error("calloc:  open_dump");
		(void)snprintf(currentdump, malsize, "%s.%010d", outfile, i);
	}
	if(real_outfile != NULL)
	{
		if(verbose > 0) log_msg(NULL, NULL, NULL,
			"Warning:  Using -v and -w together probably doesn't do what you'd expect");
		currentdump = strdup(real_outfile);
	}
	if(pd == NULL) error("Null pd\n");
	dumper = pcap_dump_open(pd, currentdump);
	if(dumper == NULL) error("%s", pcap_geterr(pd));
	pcap_userdata = (u_char *)dumper;
	free(currentdump);
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	const struct timeval *tvp;
	struct pcap_pkthdr *hh;
	FILE *f = NULL;
	int i, foo = 0;
	struct filter *filter;

	raise(SIGCHLD);
	hh = (struct pcap_pkthdr *)(char *)h;
	(const struct timeval *)tvp = &hh->ts;
	if(stop && (tvp->tv_sec > stop)) return;
	if(start && (tvp->tv_sec < start)) return;
	if(!firstsec || (firstsec > tvp->tv_sec)) firstsec = tvp->tv_sec;
	if(lastsec < tvp->tv_sec) lastsec = tvp->tv_sec;
	if(rewrite != 0) f = (FILE *)dumper;

	if(!(filter = (struct filter *)calloc(1, sizeof(struct filter))))
		error("calloc:  pcap_print_pt");
	for(i = 1; i <= nfilters; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, hh->len, hh->caplen))
		{
			if(fcount_pt[i] < hh->caplen)
				hh->caplen = fcount_pt[i];
			if((rewrite != 0) && !foo && (fcount_pt[i] > 0))
			{
				(void)fwrite((char *)hh, sizeof(*hh), 1, f);
				(void)fwrite((char *)sp, hh->caplen, 1, f);
				foo++;
			}
			if((callback_pt[i] != 0) && (callbacks != 0))
			{
				make_filter(i, filter);
				callback_handler((u_char *)sp, filter, hh->caplen, hh, 1);
			}
		}
	}
	free(filter);
	(void)handle_doctrine(hh, sp);
	if((rewrite != 0) && !foo && (default_snaplen != 0))
	{
		if((default_snaplen != DEFAULT_SNAPLEN) &&
			(default_snaplen < hh->caplen))
			hh->caplen = default_snaplen;
		(void)fwrite((char *)hh, sizeof(*hh), 1, f);
		(void)fwrite((char *)sp, hh->caplen, 1, f);
		foo++;
	}
}

