#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_doctrine.h"
#include "shoki_config.h"
#include "shoki_callback.h"
#include <gmp.h>
#include "shoki_rt.h"
#include <math.h>

#define SWAPLONG(y) \
((((y)&0xff)<<24) | (((y)&0xff00)<<8) | (((y)&0xff0000)>>8) | (((y)>>24)&0xff))

#define TCPDUMP_MAGIC 0xa1b2c3d4


int default_snaplen, start, stop, firstsec, lastsec, silent = 0;
char *currentfile, *dev, *outfile, *fname, *filterlist;
int nfilters, fcount_pt[MAX_FILTERS], callback_pt[MAX_FILTERS];
int d_pt[MAX_FILTERS], t_pt[MAX_FILTERS], q_pt[MAX_FILTERS];
int h_pt[MAX_FILTERS];
char **filter_pt, **call_args_pt, **hex_pt, **type_pt, **id_pt;
regex_t **regex_pt;
char *dumpfile = NULL;
struct bpf_program **fcode_pt;
struct bpf_program base_fcode;
int callbacks = 1;
int linux_kludge = 0, timeval_kludge = 0, fragre = 0;
int quick = 0, sortby = 0;

FILE *logfile = NULL;
int fd;
pcap_t *pd;

int pf_outformat[] = {};
char *pf_outdelim[] = {};

pcap_handler out = pcap_print_pt;

struct folist foolist;
struct dlist difflist;

int
main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL, *config = NULL;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "aAC:E:F:ikKL:pPqr:s:S:v:")) != EOF)
		switch(op)
		{
			case 'a':
				sortby = PF_SRC;
				break;
			case 'A':
				cfile = optarg;
				break;
			case 'd':
				sortby = PF_DST;
				break;
			case 'E':
				stop = atoi(optarg);
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'i':
				sortby = PF_IPID;
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'L':
				currentfile = optarg;
				break;
			case 'p':
				sortby = PF_SPORT;
				break;
			case 'P':
				sortby = PF_DPORT;
				break;
			case 'q':
				sortby = PF_SEQ;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);	
				break;
			case 'v':
				verbose = atoi(optarg);
				break;
		}

	base_filter = get_filter(&argv[optind]);

	((cfile == NULL) ? (parse_shoki_config(RT_CONFIG)) :
		(parse_shoki_config(cfile)));
	if((pf_outformat[0] == 0) || (pf_outdelim[0] == NULL))
		(void)default_formats(pf_outformat_d, pf_outdelim_d, pf_type_d);
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(filterlist == NULL) error("No filterlist defined.");
	config = read_file(filterlist);
	if(dumpfile == NULL) error("No dumpfile");
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	(void)fake_pcap(pd, 0);
	if(currentfile != NULL)
	{
		if(*currentfile == '-')
			logfile = stdout;
		else
			logfile = fopen(currentfile, "a");
	}
	if(sortby == 0) sortby  = pf_outformat[0];
	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "rt starting");
	(void)read_zdumps(dumpfile, out);
	(void)report();
	if(logfile) (void)fclose(logfile);
	exit(0);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	struct ip *ip;
	struct tcphdr *tp;
	struct udphdr *up;
	struct folist *fl;
	mpz_t mfoo;
	u_int hlen;
	int foo, i;
	u_int16_t sport = 0, dport = 0, seq = 0;

	if(start != 0) if(h->ts.tv_sec < start) return;
	if(stop != 0) if(h->ts.tv_sec > stop) return;
	foo = 0;
	if(nfilters == 0) error("No filters defined");
	for(i = 1; i <= nfilters && foo == 0; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
			foo++;
	}
	if(foo == 0) return;
	mpz_init(mfoo);
	mpz_set_ui(mfoo, 0);
	ip = (struct ip *)(sp + sizeof(struct ether_header));
	hlen = (ip->ip_hl * 4) - 20;
	switch(ip->ip_p)
	{
		case IPPROTO_TCP:
			tp = (struct tcphdr *)(sp + sizeof(struct ether_header)
				+ sizeof(struct ip) + hlen);
			sport = ntohs(tp->th_sport);
			dport = ntohs(tp->th_dport);
			seq = ntohl(tp->th_seq);
			break;
		case IPPROTO_UDP:
			up = (struct udphdr *)(sp + sizeof(struct ether_header)
				+ sizeof(struct ip) + hlen);
			sport = ntohs(up->uh_sport);
			dport = ntohs(up->uh_dport);
			break;
		default:
			break;
	}
	switch(sortby)
	{
		case PF_SRC:
			mpz_set_ui(mfoo, ntohl(ip->ip_src.s_addr));
			break;
		case PF_DST:
			mpz_set_ui(mfoo, ntohl(ip->ip_dst.s_addr));
			break;
		case PF_SPORT:
			mpz_set_ui(mfoo, sport);
			break;
		case PF_DPORT:
			mpz_set_ui(mfoo, dport);
			break;
		case PF_SEQ:
			mpz_set_ui(mfoo, seq);
			break;
		case PF_IPID:
			mpz_set_ui(mfoo, ntohs(ip->ip_id));
			break;
		default:
			error("Requested sort field not supported");
			break;
	}
	for(fl = &foolist; fl->next; fl = fl->next)
		if(mpz_cmp(fl->foo, mfoo) == 0) goto fdone;
	mpz_set(fl->foo, mfoo);
	if(!(fl->next = (struct folist *)calloc(1,
		sizeof(struct folist))))
		error("calloc:  pcap_print_pt");
fdone:
	mpz_clear(mfoo);
}

void
report()
{
	switch(sortby)
	{
		case PF_SRC:
		case PF_DST:
		case PF_SEQ:
			(void)birthday_spacings(32);
			break;
		case PF_SPORT:
		case PF_DPORT:
		case PF_IPID:
			(void)birthday_spacings(16);
			break;
		default:
			error("Requested sort field not supported");
			break;
	}
}

/*
This is similar to the birthday spacings test as used in Marsaglia's
DIEHARD suite of tests.  -spb
*/
void
birthday_spacings(int np)
{
	struct folist *fl, *fl2;
	struct dlist *d;
	int i = 0, j = 0, k, m = 0;
	int count[MAX_BIGHASH_SIZE];
	u_int32_t min = 65536, max = 0, diff = 0;
	u_int64_t n = 0;
	double lambda, chi = 0;
	mpf_t P, e, f, p, mlambda, mtmp, mchi;

	mpf_set_default_prec(64);
	mpf_init(f);
	mpf_init(e);
	mpf_init(p);
	mpf_init(mlambda);
	mpf_init(mtmp);
	mpf_init(mchi);
	mpf_init(P);
	log_msg(NULL, NULL, NULL, "Birthday Spacings Test:");
	if(np < 18)
		log_msg(NULL, NULL, NULL, "WARNING:  Problem space (2^%d) is smaller than recommended ( >= 2^18)", np);
	for(fl = &foolist; fl->next; fl = fl->next)
	{
		fl->index = m;
		if(min == 65536) min = mpz_get_ui(fl->foo);
		switch(np)
		{
			case 32:
				if(mpz_get_ui(fl->foo) < min)
					min = mpz_get_ui(fl->foo);
				if(mpz_get_ui(fl->foo) > max)
					max = mpz_get_ui(fl->foo);
				break;
			case 16:
				if(mpz_get_ui(fl->foo) < min)
					min = mpz_get_ui(fl->foo);
				if(mpz_get_ui(fl->foo) > max) 
					max = mpz_get_ui(fl->foo);
				break;
		}
		m++;
	}
	diff = max - min;
	log_msg(NULL, NULL, NULL, "Population:  %d", m);
	if(m < MIN_POPULATION) error("Population too small;  aborting.");
	log_msg(NULL, NULL, NULL, "Spread:  %u", diff);
	n = pow(2, np);
	lambda = pow(m, 3) / (double)(4 * n);
	log_msg(NULL, NULL, NULL, "lambda:  %15.12f", lambda);
	if(lambda > MAX_LAMBDA)
		error("Lambda is too large;  aborting.");
	for(j = 0; j <= m; j++)
		count[j] = 0;
	for(fl = &foolist, k = 0; fl->next; fl = fl->next, k++)
	{
		if((verbose > 1) && ((k % (m / 10)) == 0))
			log_msg(NULL, NULL, NULL, "Passes left:  %d", (m - k));
		for(d = &difflist; d->next; d = d->next)
		{
			d->count = 0;
			d->diff = 0;
		}
		j = 0;
		for(fl2 = &foolist; fl2->next; fl2 = fl2->next)
		{
			if(fl->index <= fl2->index) continue;
			if(fl->foo == fl2->foo) continue;
			i = abs(fl->foo - fl2->foo);
			for(d = &difflist; d->next && (d->diff != 0) &&
				(d->diff != i); d = d->next);
			if((d->diff != i) || (d->diff == 0))
			{
				d->diff = i;
				d->count = 0;
				if(!(d->next = (struct dlist *)calloc(1,
					sizeof(struct dlist))))
					error("calloc:  report");
			}
			d->count++;
			if(d->count == 2)
				j++;
		}
		count[j]++;
	}
	mpf_set_d(mlambda, lambda);
	mexp(e, mlambda);
	mpf_ui_div(e, 1, e);
	mpf_set_d(f, 1.0);
	for(i = 0; i <= m; i++)
	{
		mpf_pow_ui(p, mlambda, i);
		if(i != 0) mpf_mul_ui(f, f, i);
		mpf_mul(P, e, p);
		mpf_div(P, P, f);
		if((verbose) && ((mpf_get_d(P) > 0) ||
			(((float)count[i]/(float)m) > 0)))
			log_msg(NULL, NULL, NULL,
			"count[%d]:  %d     count/m:  %f     P[%d]:  %f",
			i, count[i], ((float)count[i]/(float)m), i,
			mpf_get_d(P));
		mpf_set_ui(mtmp, count[i]);
		mpf_div_ui(mtmp, mtmp, m);
		mpf_sub(mtmp, P, mtmp);
		mpf_pow_ui(mtmp, mtmp, 2);
		mpf_div(mtmp, mtmp, P);
		mpf_add(mchi, mchi, mtmp);
	}
	chi = mpf_get_d(mchi);
	log_msg(NULL, NULL, NULL, "Chi-square:  %f", chi);
	mpf_clear(f);
	mpf_clear(e);
	mpf_clear(p);
	mpf_clear(mlambda);
	mpf_clear(mtmp);
	mpf_clear(mchi);
	mpf_clear(P);
}


/* Approximate e^x for noninteger values of x */
void
mexp(mpf_t foo, mpf_t x)
{
	u_int64_t n;
	mpf_t mtmp, f;

	mpf_set_default_prec(64);
	mpf_init(mtmp);
	mpf_init(f);
	mpf_set_ui(f, 1);
	mpf_set_ui(foo, 1);
	for(n = 1; n <= MAX_FACTORIAL; n++)
	{
		mpf_mul_ui(f, f, n);
		mpf_pow_ui(mtmp, x, n);
		mpf_div(mtmp, mtmp, f);
		mpf_add(foo, foo, mtmp);
	}
	mpf_clear(mtmp);
	mpf_clear(f);
	return;
}
